@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
 dynafig
 Dynamic matplotlib in jupyer notebooks
@@ -155,7 +154,7 @@ class DynamicFig(object):
                     https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.figure.html#
                 It is recommended not to use 'figsize'
         """                
-        self.hdisplay   = display.display("", display_id=True)
+        self.hdisplay   = display.display(display_id=True)
         self.axes       = []
         self.fig        = None
         self.row_size   = int(row_size)
@@ -208,11 +207,19 @@ class DynamicFig(object):
         self.this_col = 0
         self.this_row = self.this_row + 1
             
-    def render(self):
+    def render(self, experimental_mode = "hdisplay"):
         """
         Plot all axes.
         Once called, no further plots can be added, but the plots can
         be updated in place
+        
+        Parameters
+        ----------
+            experimental_mode : str, optional
+                How to render an updated graph.
+                The default, hdisplay has worked on JupyterHub but some users
+                reported that this is not working for Jupyter.
+                Use 'canvas' in this case.
         """
         _log.verify( not self.hdisplay is None, "Cannot call render() after close() was called")        
         if self.this_row == 0 and self.this_col == 0:
@@ -234,22 +241,35 @@ class DynamicFig(object):
                 catch.execute( self.fig )
             self.caught = []
             # close.
-            # This removes a second shaddow draw in Jupyter            
-        self.hdisplay.update(self.fig)  
+            # This removes a second shaddow draw in Jupyter       
+        if experimental_mode == "hdisplay":
+            self.hdisplay.update(self.fig)  
+        else:
+            _log.verify( experimental_mode == "canvas", "'experimental_mode' must be 'hdisplay' or 'canvas'")
+            self.fig.canvas.draw()
         
-    def close(self):
+    def close(self, experimental_mode = "hdisplay"):
         """
         Close down the figure. Does not clear the figure.
         Call this to remove the resiudal print in jupyter at the end of your animation
+
+        Parameters
+        ----------
+            experimental_mode : str, optional
+                Passed on to render()
+                How to render an updated graph.
+                The default, hdisplay has worked on JupyterHub but some users
+                reported that this is not working for Jupyter.
+                Use 'canvas' in this case.
         """
         if not self.hdisplay is None:
-            self.render()            
+            self.render(experimental_mode)            
             plt.close(self.fig)  
         self.hdisplay = None
 
     def __getattr__(self, key): 
         """
-        This features allows delaying a call to figure to when it is constructed.
+        Allows delaying a call to figure to when it is constructed.
         E.g. if you call self.tight_layout() before render() then tight_layout() will called after render().        
         This will not work with functions which return anything
         """
@@ -431,4 +451,3 @@ def colors_xkcd():
 
 
 
-
